@@ -1,4 +1,4 @@
-require File.join(File.dirname(__FILE__), "spec_helper")
+require File.join(File.dirname(File.expand_path(__FILE__)), "spec_helper")
 
 describe "Sequel::Plugins::ValidationHelpers" do
   before do
@@ -210,6 +210,20 @@ describe "Sequel::Plugins::ValidationHelpers" do
     @m.should be_valid
     @m.value = '123456'
     @m.should_not be_valid
+    @m.errors[:value].should == ['is longer than 5 characters']
+    @m.value = nil
+    @m.should_not be_valid
+    @m.errors[:value].should == ['is not present']
+  end
+
+  specify "should support validates_max_length with nil value" do
+    @c.set_validations{validates_max_length(5, :value, :message=>'tl', :nil_message=>'np')}
+    @m.value = '123456'
+    @m.should_not be_valid
+    @m.errors[:value].should == ['tl']
+    @m.value = nil
+    @m.should_not be_valid
+    @m.errors[:value].should == ['np']
   end
 
   specify "should support validates_min_length" do
@@ -258,6 +272,29 @@ describe "Sequel::Plugins::ValidationHelpers" do
     @m.should be_valid
     @m.value = '.0123'
   end
+  
+  specify "should support validates_type" do
+    @c.set_validations{validates_type(Integer, :value)}
+    @m.value = 123
+    @m.should be_valid
+    @m.value = '123'
+    @m.should_not be_valid
+    @m.errors.full_messages.should == ['value is not a Integer']
+    
+    @c.set_validations{validates_type(:String, :value)}
+    @m.value = '123'
+    @m.should be_valid
+    @m.value = 123
+    @m.should_not be_valid
+    @m.errors.full_messages.should == ['value is not a String']
+    
+    @c.set_validations{validates_type('Integer', :value)}
+    @m.value = 123
+    @m.should be_valid
+    @m.value = 123.05
+    @m.should_not be_valid
+    @m.errors.full_messages.should == ['value is not a Integer']
+  end
 
   specify "should support validates_presence" do
     @c.set_validations{validates_presence(:value)}
@@ -280,18 +317,14 @@ describe "Sequel::Plugins::ValidationHelpers" do
     @c.columns(:id, :username, :password)
     @c.set_dataset MODEL_DB[:items]
     @c.set_validations{validates_unique(:username)}
-    @c.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        
-        case sql
-        when /COUNT.*username = '0records'/
-          yield({:v => 0})
-        when /COUNT.*username = '1record'/
-          yield({:v => 1})
-        end
+    @c.dataset._fetch = proc do |sql|
+      case sql
+      when /COUNT.*username = '0records'/
+        {:v => 0}
+      when /COUNT.*username = '1record'/
+        {:v => 1}
       end
-    })
+    end
     
     @user = @c.new(:username => "0records", :password => "anothertest")
     @user.should be_valid
@@ -322,18 +355,14 @@ describe "Sequel::Plugins::ValidationHelpers" do
     @c.columns(:id, :username, :password)
     @c.set_dataset MODEL_DB[:items]
     @c.set_validations{validates_unique([:username, :password])}
-    @c.dataset.extend(Module.new {
-      def fetch_rows(sql)
-        @db << sql
-        
-        case sql
-        when /COUNT.*username = '0records'/
-          yield({:v => 0})
-        when /COUNT.*username = '1record'/
-          yield({:v => 1})
-        end
+    @c.dataset._fetch = proc do |sql|
+      case sql
+      when /COUNT.*username = '0records'/
+        {:v => 0}
+      when /COUNT.*username = '1record'/
+        {:v => 1}
       end
-    })
+    end
     
     @user = @c.new(:username => "0records", :password => "anothertest")
     @user.should be_valid
@@ -354,7 +383,7 @@ describe "Sequel::Plugins::ValidationHelpers" do
 
     @user = @c.load(:id=>1, :username => "0records", :password => "anothertest")
     @user.should be_valid
-    MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE (((username = '0records') AND (password = 'anothertest')) AND (id != 1)) LIMIT 1"
+    MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND (password = 'anothertest') AND (id != 1)) LIMIT 1"
     @user = @c.new(:username => "0records", :password => "anothertest")
     @user.should be_valid
     MODEL_DB.sqls.last.should == "SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND (password = 'anothertest')) LIMIT 1"
@@ -364,17 +393,41 @@ describe "Sequel::Plugins::ValidationHelpers" do
     @c.columns(:id, :username, :password)
     @c.set_dataset MODEL_DB[:items]
     @c.set_validations{validates_unique(:username){|ds| ds.filter(:active)}}
-    @c.dataset.extend(Module.new {
-      def fetch_rows (sql)
-        @db << sql
-        yield({:v => 0})
-      end
-    })
+    @c.dataset._fetch = {:v=>0}
     
     MODEL_DB.reset
     @c.new(:username => "0records", :password => "anothertest").should be_valid
     @c.load(:id=>3, :username => "0records", :password => "anothertest").should be_valid
     MODEL_DB.sqls.should == ["SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND active) LIMIT 1",
-                    "SELECT COUNT(*) AS count FROM items WHERE (((username = '0records') AND active) AND (id != 3)) LIMIT 1"]
+                    "SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND active AND (id != 3)) LIMIT 1"]
+  end
+
+  it "should support :only_if_modified option for validates_unique, and not check uniqueness for existing records if values haven't changed" do
+    @c.columns(:id, :username, :password)
+    @c.set_dataset MODEL_DB[:items]
+    @c.set_validations{validates_unique([:username, :password], :only_if_modified=>true)}
+    @c.dataset._fetch = {:v=>0}
+    
+    MODEL_DB.reset
+    @c.new(:username => "0records", :password => "anothertest").should be_valid
+    MODEL_DB.sqls.should == ["SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND (password = 'anothertest')) LIMIT 1"]
+    MODEL_DB.reset
+    m = @c.load(:id=>3, :username => "0records", :password => "anothertest")
+    m.should be_valid
+    MODEL_DB.sqls.should == []
+
+    m.username = '1'
+    m.should be_valid
+    MODEL_DB.sqls.should == ["SELECT COUNT(*) AS count FROM items WHERE ((username = '1') AND (password = 'anothertest') AND (id != 3)) LIMIT 1"]
+
+    m = @c.load(:id=>3, :username => "0records", :password => "anothertest")
+    MODEL_DB.reset
+    m.password = '1'
+    m.should be_valid
+    MODEL_DB.sqls.should == ["SELECT COUNT(*) AS count FROM items WHERE ((username = '0records') AND (password = '1') AND (id != 3)) LIMIT 1"]
+    MODEL_DB.reset
+    m.username = '2'
+    m.should be_valid
+    MODEL_DB.sqls.should == ["SELECT COUNT(*) AS count FROM items WHERE ((username = '2') AND (password = '1') AND (id != 3)) LIMIT 1"]
   end
 end 
